(* ocamlopt unix.cmxa nondet.ml permutations.ml -o permutations *)

module Permutations (M:Nondet.S) = struct

  open Nondet.Sugar(M)

  let rec pick acc = function
    | [] -> fail
    | hd::tl ->
        orelse
          (return (hd, List.rev_append acc tl))
          (pick (hd::acc) tl)

  let rec permutation l =
    if l = [] then return l else
    let* e,l' = pick [] l in
    let* perm = permutation l' in
    return (e::perm)

  let run size =
    let c = ref 0 in
    iter
      (permutation (Array.to_list (Array.init size (fun i -> i))))
      (fun _ -> incr c) ;
    !c

end

let rec fact acc n = if n = 1 then acc else fact (acc*n) (n-1)
let fact n = fact 1 n

let () =
  let module ML = Permutations(Nondet.List) in
  let module MC = Permutations(Nondet.Cont) in
  for len = 8 to 12 do
    Printf.printf "\n--- %d ---\n%!" len ;
    let fn = fact len in
    (* Continuations *)
    let t0 = Unix.gettimeofday () in
    let n = MC.run len in
    let t1 = Unix.gettimeofday () in
    Printf.printf "cont: %.2fs\n%!" (t1-.t0) ;
    assert (n = fn) ;
    (* Lists *)
    let t0 = Unix.gettimeofday () in
    let n = ML.run len in
    let t1 = Unix.gettimeofday () in
    Printf.printf "list: %.2fs\n%!" (t1-.t0) ;
    assert (n = fn) ;
  done
